import numpy as np
import onnxruntime


class AVESOnnxModel:
    """Wrapper class for ONNX models."""

    def __init__(self, model_path: str):
        self.ort_session = onnxruntime.InferenceSession(model_path)

    def __call__(self, inputs: np.ndarray) -> np.ndarray:
        inputs = {self.ort_session.get_inputs()[0].name: inputs}
        outputs = self.ort_session.run(None, inputs)
        return outputs[0]



"""Unit tests for the aves package.

IMPORTANT! This assumes that you have downloaded the checkpoint:
1. birdaves-biox-large.onnx

and placed them in the ../models folder.
"""

import os
import subprocess
from pathlib import Path

import numpy as np
import torch

from aves import AVESClassifier, AVESOnnxModel, load_feature_extractor


def test_feature_extractor_loader():
    # Test loading feature extractor
    model = load_feature_extractor(
        config_path="config/default_cfg_aves-base-all.json", device="cpu", for_inference=True
    )
    assert model is not None
    assert model.config.get("encoder_embed_dim") == 768


def test_aves_feature_extractor():
    # Test AVES feature extractor
    model = load_feature_extractor(
        config_path="config/default_cfg_aves-base-all.json", device="cpu", for_inference=True
    )
    embeddings = model.extract_features(torch.rand(2, 16000))

    assert embeddings.shape == (2, 49, 768)

    # test multiple layers
    embeddings = model.extract_features(torch.rand(2, 16000), layers=[-2, -1])
    assert len(embeddings) == 2
    assert embeddings[0].shape == (2, 49, 768)
    assert embeddings[1].shape == (2, 49, 768)


def test_birdaves_feature_extractor():
    # Test BirdAVES feature extractor
    model = load_feature_extractor(
        config_path="config/default_cfg_birdaves-biox-large.json", device="cpu", for_inference=True
    )
    embeddings = model.extract_features(torch.rand(2, 16000))
    print("Embeddings shape", embeddings.shape)
    assert embeddings.shape == (2, 49, 1024)


def test_aves_classifier():
    # Initialize an AVES classifier with 10 target classes
    print("Initializing AVES classifier...")
    model = AVESClassifier(
        config_path="config/default_cfg_birdaves-biox-large.json",
        model_path=None,
        num_classes=10,
        for_inference=True,
        freeze_feature_extractor=True,
        device="cpu",
    )

    print("Running forward pass...")
    # Create a 1-second random sound
    x = torch.rand(2, 16000)
    y = torch.tensor([0, 1])

    # Run the forward pass
    loss, logits = model(inputs=x, labels=y)

    assert loss is not None
    assert logits.shape[0] == 2
    assert logits.shape[1] == 10


def test_aves_onnx():
    # Initialize an AVES ONNX model
    # check if the model is downloaded
    if not Path("models/birdaves-biox-large.onnx").exists():
        print("Please download the birdaves-biox-large.onnx model file to the models/ folder, skipping test...")
        return
    model = AVESOnnxModel("aves/birdaves-biox-large.onnx")
    # Create two 1-second random sounds
    x = torch.rand(2, 16000, dtype=torch.float32).numpy()
    # Run the forward pass
    outputs = model(x)
    assert outputs.shape == (2, 49, 1024)


def test_cli():
    # Test the AVES CLI
    p = subprocess.run(
        [
            "aves",
            "-c",
            "config/default_cfg_birdaves-biox-large.json",
            "--path_to_audio_dir",
            "example_audios/",
            "--output_dir",
            "example_audios/",
            "--save_as",
            "npy",
            "--device",
            "cpu",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=True,
    )

    assert p.returncode == 0
    assert b"Processing 2 audio files..." in p.stdout
    assert b"Saving embedding to example_audios/" in p.stdout

    # check that embdding files are saved
    assert (Path("example_audios") / "XC448414 - Eurasian Bullfinch - Pyrrhula pyrrhula.embedding.npy").exists()
    assert (Path("example_audios") / "XC936872 - Helmeted Guineafowl - Numida meleagris.embedding.npy").exists()

    # load one of them
    emb = np.load("example_audios/XC448414 - Eurasian Bullfinch - Pyrrhula pyrrhula.embedding.npy", allow_pickle=True)
    assert isinstance(emb, np.ndarray)
    assert emb.shape == (2, 1049, 1024)

    # remove these files
    os.remove("example_audios/XC448414 - Eurasian Bullfinch - Pyrrhula pyrrhula.embedding.npy")
    os.remove("example_audios/XC936872 - Helmeted Guineafowl - Numida meleagris.embedding.npy")


def test_cli_multiple_layers():
    # Test the AVES CLI
    p = subprocess.run(
        [
            "aves",
            "-c",
            "config/default_cfg_birdaves-biox-large.json",
            "--path_to_audio_dir",
            "example_audios/",
            "--output_dir",
            "example_audios/",
            "--layers",
            "0-3",
            "--save_as",
            "npy",
            "--device",
            "cpu",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=True,
    )

    assert p.returncode == 0
    assert b"Processing 2 audio files..." in p.stdout
    assert b"Saving embedding to example_audios/" in p.stdout

    # check that embdding files are saved
    assert (Path("example_audios") / "XC448414 - Eurasian Bullfinch - Pyrrhula pyrrhula.embedding.npy").exists()
    assert (Path("example_audios") / "XC936872 - Helmeted Guineafowl - Numida meleagris.embedding.npy").exists()

    # load one of them
    emb = np.load("example_audios/XC448414 - Eurasian Bullfinch - Pyrrhula pyrrhula.embedding.npy", allow_pickle=True)
    assert len(emb) == 4
    assert isinstance(emb[0], np.ndarray)
    assert emb[0].shape == (2, 1049, 1024)

    # remove these files
    os.remove("example_audios/XC448414 - Eurasian Bullfinch - Pyrrhula pyrrhula.embedding.npy")
    os.remove("example_audios/XC936872 - Helmeted Guineafowl - Numida meleagris.embedding.npy")


def test_cli_multiple_layers_2():
    # Test the AVES CLI
    p = subprocess.run(
        [
            "aves",
            "-c",
            "config/default_cfg_birdaves-biox-large.json",
            "--path_to_audio_dir",
            "example_audios/",
            "--output_dir",
            "example_audios/",
            "--layers",
            "0,1,11",
            "--save_as",
            "npy",
            "--device",
            "cpu",
        ],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        check=True,
    )

    assert p.returncode == 0
    assert b"Processing 2 audio files..." in p.stdout
    assert b"Saving embedding to example_audios/" in p.stdout

    # check that embdding files are saved
    assert (Path("example_audios") / "XC448414 - Eurasian Bullfinch - Pyrrhula pyrrhula.embedding.npy").exists()
    assert (Path("example_audios") / "XC936872 - Helmeted Guineafowl - Numida meleagris.embedding.npy").exists()

    # load one of them
    emb = np.load("example_audios/XC448414 - Eurasian Bullfinch - Pyrrhula pyrrhula.embedding.npy", allow_pickle=True)
    assert len(emb) == 3
    assert isinstance(emb[0], np.ndarray)
    assert emb[0].shape == (2, 1049, 1024)

    # remove these files
    os.remove("example_audios/XC448414 - Eurasian Bullfinch - Pyrrhula pyrrhula.embedding.npy")
    os.remove("example_audios/XC936872 - Helmeted Guineafowl - Numida meleagris.embedding.npy")


# THIS TEST FAILS BECAUSE OF ONNX OUTPUT DIFFERENCES
# def test_onnx_vs_torchaudio_output():
#     # Initialize an AVES ONNX model
#     onnx_model = AVESOnnxModel("aves/birdaves-biox-large.onnx")

#     # BirdAVES feature extractor
#     model = load_feature_extractor(
#         config_path="config/default_cfg_birdaves-biox-large.json",
#         model_path="aves/birdaves-biox-large.torchaudio.pt",
#         device="cpu",
#         for_inference=True,
#     )

#     # Create a 1-second random sound
#     x = torch.rand(2, 16000, dtype=torch.float32).numpy()

#     # Run the forward pass
#     onnx_outputs = onnx_model(x)
#     torchaudio_outputs = model.extract_features(torch.from_numpy(x))

#     assert onnx_outputs.shape == torchaudio_outputs.shape
#     # !!Fails to produce identical outputs!!
#     assert torch.allclose(torch.from_numpy(onnx_outputs), torchaudio_outputs, atol=1e-5)



"""AVES package for audio feature extraction."""

from .aves import (
    AVESTorchaudioWrapper,
    AVESClassifier,
    load_feature_extractor,
)
from .aves_onnx import AVESOnnxModel

__all__ = [
    "AVESTorchaudioWrapper",
    "AVESClassifier",
    "load_feature_extractor",
    "AVESOnnxModel",
]



"""Utility functions for Aves."""

import os
from pathlib import Path

import numpy as np
import torch
import torchaudio

ALLOWED_AUDIO_FILE_EXTENSIONS = ["wav", "mp3", "flac", "ogg", "m4a"]
TARGET_SR = 16000  # Aves works with 16kHz audio


def parse_audio_file_paths(audio_dir: str, audio_file_extension: str | None = None) -> list[Path]:
    """Parse audio file paths from a directory or a list of paths.

    Args:
        audio_dir (str): Path to the directory containing the audio files.
        audio_file_extension (str, optional): Extension of the audio files to process. Defaults to None.

    Returns:
        list[Path]: List of Path objects for the audio files.
    """
    audio_dir = Path(audio_dir)
    if audio_file_extension is None:
        audio_files = []
        print(f"Searching for audio files in {audio_dir} with extensions: {ALLOWED_AUDIO_FILE_EXTENSIONS}")
        for ext in ALLOWED_AUDIO_FILE_EXTENSIONS:
            audio_files.extend(list(audio_dir.glob(f"*.{ext}")))

        print(f"Fetched {len(audio_files)} audio files from {audio_dir}")
    else:
        # check if the provided extension is supported
        if audio_file_extension not in ALLOWED_AUDIO_FILE_EXTENSIONS:
            raise ValueError(
                f"Unsupported audio file extension: {audio_file_extension}. Supported extensions are: {ALLOWED_AUDIO_FILE_EXTENSIONS}"
            )
        audio_files = list(audio_dir.glob(f"*.{audio_file_extension}"))

    return audio_files


def save_embedding(embedding: torch.Tensor | list[torch.Tensor], output_file: Path, save_as: str) -> None:
    """Save the embedding to a file.

    Args:
        embedding (torch.Tensor): The embedding to save.
        output_file (Path): Path to save the embedding.
        save_as (str): Format to save the embedding, either 'pt' or 'npy'.
    """

    if save_as == "pt":
        torch.save(embedding, output_file + ".pt")

    elif save_as == "npy":
        if isinstance(embedding, list):
            np_embedding = [e.cpu().numpy() for e in embedding]
        else:
            np_embedding = embedding.cpu().numpy()

        # numpy adds extension .npy automatically
        np.save(output_file, np_embedding)

    else:
        raise ValueError(f"Unsupported save_as format: {save_as}")

    print(f"Saving embedding to {output_file}...")


def load_audio(audio_file: str | os.PathLike | Path, mono: bool = False, mono_avg: bool = False) -> torch.Tensor:
    """Load audio file and resample if necessary.

    Args:
        audio_file (str | os.PathLike | Path): Path to the audio file.
        mono (bool, optional): Whether to load the audio as mono. Defaults to False.
            If True, the audio will be converted to mono by averaging the channels.
        mono_avg (bool, optional): Whether to average the channels for mono conversion. Defaults to False.
            False = keep the first channel.

    Returns:
        audio (torch.Tensor): Audio tensor
    """
    audio, sr = torchaudio.load(audio_file)
    if sr != TARGET_SR:
        print(f"CAUTION: Resampling audio from {sr} to {TARGET_SR} required by AVES models...")
        resampler = torchaudio.transforms.Resample(sr, TARGET_SR)
        audio = resampler(audio)

    if mono and audio.shape[0] > 1:
        if mono_avg:
            audio = audio.mean(dim=0, keepdim=False)
        else:
            audio = audio[0]

    return audio



"""AVES: Audio and Visual Event Detection with Self-supervised learning"""

import json
from pathlib import Path

import torch
import torch.nn as nn
from torchaudio.models import wav2vec2_model


def load_config(config_path: str) -> dict:
    with open(config_path, "r") as ff:
        obj = json.load(ff)
    return obj


class AVESTorchaudioWrapper(nn.Module):
    def __init__(self, config_path: str | Path, model_path: str | Path = None, device: str = "cuda"):
        super().__init__()

        self.config = load_config(str(config_path))

        print("Loading Hubert model (without AVES weights)...")
        self.model = wav2vec2_model(**self.config, aux_num_out=None)
        if model_path is not None:
            print("Loading AVES model weights from", model_path)
            self.model.load_state_dict(torch.load(str(model_path), weights_only=True))

        self.device = device

    def forward(self, inputs: torch.Tensor, layers: list[int] | int | None = -1) -> torch.Tensor | list[torch.Tensor]:
        """For training, use the forward method to get the output of the model.

        Args:
            inputs (torch.Tensor): Input tensor
            layers (list[int] | int | None, optional): Layer(s) to extract features from. Defaults to -1 (last layer).

        Returns:
            torch.Tensor | list[torch.Tensor]: Output tensor(s) from the model
        """
        out = self.model.extract_features(inputs)[0]

        if layers is not None and isinstance(layers, int):
            return out[layers]

        if layers and isinstance(layers, list):
            return [out[layer] for layer in layers]

        # return all layers
        return out

    @torch.no_grad()
    def extract_features(
        self,
        inputs: torch.Tensor,
        layers: list[int] | int | None = -1,
    ) -> torch.Tensor:
        """For inference, use the extract_features method to get the output of the model.

        Args:
            inputs (torch.Tensor): Input tensor
            layers (list[int] | int | None, optional): Layer(s) to extract features from. Defaults to -1 (last layer).

        Returns:
            torch.Tensor | list[torch.Tensor]: Output tensor
        """
        if inputs.ndim == 1:
            inputs = inputs.unsqueeze(0)
        return self.forward(inputs, layers)


def load_feature_extractor(
    config_path: str | Path, model_path: str | Path = None, device: str = "cuda", for_inference: bool = True
) -> AVESTorchaudioWrapper:
    """Load the AVES feature extractor model

    Args:
        config_path (str | Path): Path to the model configuration file
        model_path (str | Path): Path to the model weights file
        device (str, optional): Device to run the model on. Defaults to "cuda".
        for_inference (bool, optional): Whether to set the underlying feature extractor to inference mode. Defaults to True.

    Returns:
        AVESTorchaudioWrapper: The AVES feature extractor model
    """
    device = "cuda" if torch.cuda.is_available() and device == "cuda" else "cpu"
    if for_inference:
        return AVESTorchaudioWrapper(config_path, model_path, device).to(device).eval()

    return AVESTorchaudioWrapper(config_path, model_path, device).to(device)


class AVESClassifier(nn.Module):
    """A classifier model using AVES as a feature extractor

    Args:
        config_path (str | Path): Path to the model configuration file
        model_path (str | Path): Path to the model weights file
        num_classes (int): Number of target classes
        freeze_feature_extractor (bool, optional): Whether to freeze the feature extractor. Defaults to True.
        for_inference (bool, optional): Whether to set the underlying feature extractor to inference mode. Defaults to False.
    """

    def __init__(
        self,
        config_path: str | Path,
        num_classes: int,
        model_path: str | Path = None,  # or a pre-trained model path
        freeze_feature_extractor: bool = True,
        for_inference: bool = False,
        device: str = "cuda",
    ):
        super().__init__()

        self.model = load_feature_extractor(config_path, model_path, for_inference=for_inference, device=device)
        embeddings_dim = self.model.config.get("encoder_embed_dim", 768)
        self.head = nn.Linear(in_features=embeddings_dim, out_features=num_classes)

        device = "cuda" if torch.cuda.is_available() and device == "cuda" else "cpu"
        self.device = device  # you still have to move the model to the device you want to use
        self.head.to(device)

        if freeze_feature_extractor:
            print("Freezing feature extractor, it will NOT be updated during training!")
            self.model.requires_grad_(False)

        if num_classes == 1:
            self.loss_func = nn.BCEWithLogitsLoss()
        else:
            self.loss_func = nn.CrossEntropyLoss()

    def forward(self, inputs: torch.Tensor, labels: torch.Tensor = None) -> tuple[torch.Tensor | None, torch.Tensor]:
        out = self.model.forward(inputs, layers=-1)
        out = out.mean(dim=1)  # mean pooling over time dimension
        logits = self.head(out)

        loss = None
        if labels is not None:
            loss = self.loss_func(logits, labels)

        return loss, logits


if __name__ == "__main__":
    # test

    # Initialize an AVES classifier with 10 target classes
    print("Initializing AVES classifier to CPU...")
    model = AVESClassifier(
        config_path="../config/birdaves-biox-large.json",
        model_path="birdaves-biox-large.torchaudio.pt",
        num_classes=10,
        for_inference=True,
        freeze_feature_extractor=True,
        device="cpu",
    )

    print("Running forward pass...")
    # Create a batch of 2 1-second random sound
    x = torch.rand(2, 16000)
    y = torch.tensor([0, 1])

    # Run the forward pass
    loss, logits = model(inputs=x, labels=y)

    print("Loss:", loss)
    print("Logits shape = ", logits.shape)



"""Main entry point for the aves package."""

import argparse
from pathlib import Path

from .aves import load_feature_extractor
from .utils import load_audio, parse_audio_file_paths, save_embedding


def parse_layers_argument(layers: str) -> list[int] | int | None:
    """Parse the layers argument from the command line

    Args:
        layers (str): Layers argument from the command line

    Returns:
        list[int] | int | None: List of layers to extract features from
    """

    if layers == "all":
        return None

    try:
        layers = int(layers)
        return layers
    except ValueError:
        # not a single integer
        pass

    # comma separated list ?
    try:
        layers = [int(layer) for layer in layers.split(",")]
        return layers
    except ValueError:
        # not a comma separated list
        pass

    # range ?
    try:
        layers = [int(layer) for layer in layers.split("-")]
        return [i for i in range(layers[0], layers[1] + 1)]
    except ValueError:
        raise ValueError("Invalid layers argument, see --help for more information")


def main():
    """Run the AVES model on a set of audio file paths"""

    parser = argparse.ArgumentParser(
        description="""Run the AVES feature extractor model on a set of audio file paths."""
    )

    parser.add_argument("-c", "--config_path", type=str, required=True, help="Path to the model configuration file")

    parser.add_argument(
        "-m",
        "--model_path",
        type=str,
        default=None,
        help="Path to the model weights file. If not provided, will be using the Hubert Model without AVES weights!!",
    )

    parser.add_argument(
        "-a", "--audio_paths", type=str, default=None, nargs="+", help="Paths to the audio files to process"
    )

    parser.add_argument(
        "--path_to_audio_dir",
        type=str,
        default=None,
        help="Path to the directory containing the audio files to process",
    )

    parser.add_argument(
        "--audio_file_extension", type=str, default=None, help="Extension of the audio files to process"
    )

    parser.add_argument(
        "--layers",
        type=str,
        default="-1",
        help="""Layers to extract features from.
        You can either say "all" for all layers, or provide a range like "0-3" for layers 0,1,2,3 (inclusive),
        or a single layer like "2" for layer 3 (starting from 0), or a comma-separated list like "0,2,4"
        for layers 0, 2, and 4.
        Default is the last layer ("-1").
        """,
    )

    parser.add_argument("--device", type=str, default="cuda", help="Device to run the model on (default: cuda)")

    parser.add_argument("--output_dir", type=str, default=".", help="Directory to save the output embedding files")

    parser.add_argument(
        "--save_as",
        type=str,
        default="pt",
        help="Format to save the output embeddings, either 'pt' or 'npy'",
    )

    args = parser.parse_args()

    if not args.model_path:
        print("---CAUTION: Running the model without AVES weights!!---")

    model = load_feature_extractor(args.config_path, args.model_path, args.device, for_inference=True)

    # if audio_dir fetch all audio files in the directory with extention audio_file_extension
    if args.audio_paths is None:
        assert args.path_to_audio_dir is not None, "Either audio_paths or path_to_audio_dir must be provided"
        audio_files = parse_audio_file_paths(args.path_to_audio_dir, args.audio_file_extension)
    else:
        audio_files = [Path(audio_path) for audio_path in args.audio_paths]

    # parse layers argument
    layers = parse_layers_argument(args.layers)

    print(f"Processing {len(audio_files)} audio files...")
    for audio_file in audio_files:
        print(f"==== Processing {audio_file} ====")

        audio = load_audio(audio_file)
        embedding = model.extract_features(audio.to(args.device), layers)
        save_embedding(embedding, Path(args.output_dir) / f"{audio_file.stem}.embedding", args.save_as)
